@@ -186,9 +186,9 @@ class ServiceDiscovery(StateEVCC):
         await self.select_energy_transfer_mode()
 
         charge_service: ChargeService = service_discovery_res.charge_service
-        offered_energy_modes: List[
-            EnergyTransferModeEnum
-        ] = charge_service.supported_energy_transfer_mode.energy_modes
+        offered_energy_modes: List[EnergyTransferModeEnum] = (
+            charge_service.supported_energy_transfer_mode.energy_modes
+        )
 
         if self.comm_session.selected_energy_mode not in offered_energy_modes:
             self.stop_state_machine(
@@ -770,9 +770,8 @@ class ChargeParameterDiscovery(StateEVCC):
             ) = await ev_controller.process_sa_schedules_v2(
                 charge_params_res.sa_schedule_list.schedule_tuples
             )
-
+            await self.comm_session.ev_controller.enable_charging(True)
             if self.comm_session.selected_charging_type_is_ac:
-
                 power_delivery_req = PowerDeliveryReq(
                     charge_progress=charge_progress,
                     sa_schedule_tuple_id=schedule_id,
@@ -799,7 +798,7 @@ class ChargeParameterDiscovery(StateEVCC):
 
             self.comm_session.selected_schedule = schedule_id
 
-            # TODO Set CP state to C max. 250 ms after sending PowerDeliveryReq
+            await self.comm_session.ev_controller.enable_charging(True)
         else:
             logger.debug(
                 "SECC is still processing the proposed charging "
@@ -875,6 +874,7 @@ class PowerDelivery(StateEVCC):
                 Namespace.ISO_V2_MSG_DEF,
             )
         elif self.comm_session.charging_session_stop_v2:
+            await self.comm_session.ev_controller.enable_charging(False)
             welding_detection_req = WeldingDetectionReq(
                 dc_ev_status=await self.comm_session.ev_controller.get_dc_ev_status()
             )
@@ -1017,7 +1017,7 @@ class MeteringReceipt(StateEVCC):
             else:
                 self.create_next_message(
                     CurrentDemand,
-                    self.build_current_demand_req(),
+                    await self.build_current_demand_req(),
                     Timeouts.CHARGING_STATUS_REQ,
                     Namespace.ISO_V2_MSG_DEF,
                 )
