@@ -14,8 +14,9 @@ import asyncio
 import logging
 import socket
 from asyncio.streams import StreamReader, StreamWriter
-from typing import Dict, List, Optional, Tuple, Union
+from typing import Any, Coroutine, Dict, List, Optional, Tuple, Union
 
+from iso15118.secc.controller.ev_data import EVSessionContext15118
 from iso15118.secc.controller.interface import EVSEControllerInterface, ServiceStatus
 from iso15118.secc.failed_responses import (
     init_failed_responses_din_spec_70121,
@@ -34,6 +35,7 @@ from iso15118.shared.messages.enums import (
     ISOV2PayloadTypes,
     ISOV20PayloadTypes,
     Protocol,
+    SessionStopAction,
 )
 from iso15118.shared.messages.iso15118_2.datatypes import (
     CertificateChain as CertificateChainV2,
@@ -45,7 +47,13 @@ from iso15118.shared.messages.iso15118_2.datatypes import (
     eMAID,
 )
 from iso15118.shared.messages.iso15118_20.common_messages import ScheduleTuple
-from iso15118.shared.messages.sdp import SDPRequest, Security, create_sdp_response
+from iso15118.shared.messages.sdp import (
+    SDPRequest,
+    SDPResponse,
+    SDPResponseWireless,
+    Security,
+    create_sdp_response,
+)
 from iso15118.shared.messages.timeouts import Timeouts
 from iso15118.shared.messages.v2gtp import V2GTPMessage
 from iso15118.shared.notifications import (
@@ -71,6 +79,7 @@ class SECCCommunicationSession(V2GCommunicationSession):
         config: Config,
         evse_controller: EVSEControllerInterface,
         evse_id: str,
+        ev_session_context: Optional[EVSessionContext15118] = None,
     ):
         # Need to import here to avoid a circular import error
         # pylint: disable=import-outside-toplevel
@@ -85,6 +94,10 @@ class SECCCommunicationSession(V2GCommunicationSession):
         self.evse_controller = evse_controller
         # EVSE ID associated with this session
         self.evse_id = evse_id
+        # EV Session context associated if available.
+        self.ev_session_context: EVSessionContext15118 = (
+            ev_session_context or EVSessionContext15118()
+        )
         # The authorization option(s) offered with ServiceDiscoveryRes in
         # ISO 15118-2 and with AuthorizationSetupRes in ISO 15118-20
         self.offered_auth_options: Optional[List[AuthEnum]] = []
@@ -164,12 +177,13 @@ class CommunicationSessionHandler:
     def __init__(
         self, config: Config, codec: IEXICodec, evse_controller: EVSEControllerInterface
     ):
-
-        self.list_of_tasks = []
-        self.udp_server = None
-        self.tcp_server = None
-        self.config = config
-        self.evse_controller = evse_controller
+        self.list_of_tasks: List[Coroutine] = []
+        self.udp_server: Optional[UDPServer] = None
+        self.tcp_server: Optional[TCPServer] = None
+        self.tcp_server_handler: Optional[asyncio.Task[Any]] = None
+        self.config: Config = config
+        self.evse_controller: EVSEControllerInterface = evse_controller
+        self.udp_processor_lock: asyncio.Lock = asyncio.Lock()
 
         # List of server status events
         self.status_event_list: List[asyncio.Event] = []
@@ -179,14 +193,21 @@ class CommunicationSessionHandler:
 
         # Receiving queue for UDP or TCP packets and session
         # triggers (e.g. pause/terminate)
-        self._rcv_queue = asyncio.Queue()
+        self._rcv_queue: asyncio.Queue = asyncio.Queue()
+
+        # Store the ip of the peer
+        self._current_peer_ip: Optional[str] = None
 
         # The comm_sessions dict keys are of type str (the IPv6 address), the
         # values are a tuple containing the SECCCommunicationSession and the
         # associated ayncio.Task object (so we can cancel the task when needed)
-        self.comm_sessions: Dict[str, (SECCCommunicationSession, asyncio.Task)] = {}
+        self.comm_sessions: Dict[str, Tuple[SECCCommunicationSession, asyncio.Task]] = (
+            {}
+        )
 
-    async def start_session_handler(self, iface: str):
+    async def start_session_handler(
+        self, iface: str, start_udp_server: Optional[bool] = True
+    ):
         """
         This method is necessary, because python does not allow
         async def __init__.
@@ -194,25 +215,22 @@ class CommunicationSessionHandler:
         constructor.
         """
 
-        self.udp_server = UDPServer(self._rcv_queue, iface)
-        udp_ready_event: asyncio.Event = asyncio.Event()
-        self.status_event_list.append(udp_ready_event)
+        if start_udp_server:
+            self.udp_server = UDPServer(self._rcv_queue, iface)
+            udp_ready_event: asyncio.Event = asyncio.Event()
+            self.status_event_list.append(udp_ready_event)
+            self.list_of_tasks.append(self.udp_server.start(udp_ready_event))
+        else:
+            logger.info(f"UDP server disabled on {iface}")
 
         self.tcp_server = TCPServer(self._rcv_queue, iface)
-        tls_ready_event: asyncio.Event = asyncio.Event()
-        self.status_event_list.append(tls_ready_event)
 
-        self.list_of_tasks = [
-            self.get_from_rcv_queue(self._rcv_queue),
-            self.udp_server.start(udp_ready_event),
-            self.tcp_server.start_tls(tls_ready_event),
-            self.check_status_task(),
-        ]
-
-        if not self.config.enforce_tls:
-            tcp_ready_event: asyncio.Event = asyncio.Event()
-            self.status_event_list.append(tcp_ready_event)
-            self.list_of_tasks.append(self.tcp_server.start_no_tls(tcp_ready_event))
+        self.list_of_tasks.extend(
+            [
+                self.get_from_rcv_queue(self._rcv_queue),
+                self.check_status_task(True),
+            ]
+        )
 
         logger.info("Communication session handler started")
 
@@ -231,10 +249,11 @@ class CommunicationSessionHandler:
         while self.check_events() is False:
             await asyncio.sleep(0.01)
 
-    async def check_status_task(self) -> None:
+    async def check_status_task(self, send_status_update: bool) -> None:
         try:
             await asyncio.wait_for(self.check_ready_status(), timeout=10)
-            await self.evse_controller.set_status(ServiceStatus.READY)
+            if send_status_update:
+                await self.evse_controller.set_status(ServiceStatus.READY)
         except asyncio.TimeoutError:
             logger.error("Timeout: Servers failed to startup")
             await self.evse_controller.set_status(ServiceStatus.ERROR)
@@ -260,18 +279,30 @@ class CommunicationSessionHandler:
                 if isinstance(notification, UDPPacketNotification):
                     await self.process_incoming_udp_packet(notification)
                 elif isinstance(notification, TCPClientNotification):
+                    if self.udp_server:
+                        self.udp_server.pause_udp_server()
                     logger.info(
                         "TCP client connected, client address is "
                         f"{notification.ip_address}."
                     )
 
                     try:
-                        comm_session, task = self.comm_sessions[notification.ip_address]
-                        comm_session.resume()
+                        comm_session, _ = self.comm_sessions[notification.ip_address[0]]
+                        ev_context = comm_session.ev_session_context
+                        comm_session = SECCCommunicationSession(
+                            notification.transport,
+                            self._rcv_queue,
+                            self.config,
+                            self.evse_controller,
+                            await self.evse_controller.get_evse_id(Protocol.UNKNOWN),
+                            ev_context,
+                        )
                     except (KeyError, ConnectionResetError) as e:
                         if isinstance(e, ConnectionResetError):
                             logger.info("Can't resume session. End and start new one.")
-                            await self.end_current_session(notification.ip_address)
+                            await self.end_current_session(
+                                notification.ip_address, SessionStopAction.TERMINATE
+                            )
                         comm_session = SECCCommunicationSession(
                             notification.transport,
                             self._rcv_queue,
@@ -285,12 +316,17 @@ class CommunicationSessionHandler:
                             Timeouts.V2G_EVCC_COMMUNICATION_SETUP_TIMEOUT
                         )
                     )
-                    self.comm_sessions[notification.ip_address] = (comm_session, task)
+                    self.comm_sessions[notification.ip_address[0]] = (
+                        comm_session,
+                        task,
+                    )
+                    self._current_peer_ip = notification.ip_address
                 elif isinstance(notification, StopNotification):
                     try:
-                        await self.end_current_session(notification.peer_ip_address)
+                        await self.end_current_session(
+                            notification.peer_ip_address, notification.stop_action
+                        )
                     except KeyError:
-                        # TODO Need to check why this KeyError happens
                         pass
                 else:
                     logger.warning(
@@ -302,10 +338,81 @@ class CommunicationSessionHandler:
             finally:
                 queue.task_done()
 
-    async def end_current_session(self, peer_ip_address: str):
-        await cancel_task(self.comm_sessions[peer_ip_address][1])
-        del self.comm_sessions[peer_ip_address]
-        logger.info(f"Existing session with {peer_ip_address} ended.")
+    def close_session(self):
+        """
+        Use this method to prevent the SECC from waiting until SECC timeout is hit to
+        terminate (For example when State A/E/F has been signalled).
+        The method feed_eof() would indicate to the tcp reader that the
+        peer has closed the connection. This would in turn cause a ConnectionResetError
+         and push StopNotification() which will eventually end the session.
+        """
+        if self._current_peer_ip is None:
+            return
+        try:
+            self.comm_sessions[self._current_peer_ip][0].reader.feed_eof()
+        except Exception as e:
+            logger.warning(f"Error while indicating EOF to transport reader: {e}")
+
+    async def end_current_session(
+        self, peer_ip_address: Any, session_stop_action: SessionStopAction
+    ):
+        try:
+            await cancel_task(self.tcp_server_handler)
+            await cancel_task(self.comm_sessions[peer_ip_address[0]][1])
+        except Exception as e:
+            logger.warning(f"Unexpected error ending current session: {e}")
+        finally:
+            if session_stop_action == SessionStopAction.TERMINATE:
+                del self.comm_sessions[peer_ip_address[0]]
+            else:
+                logger.info(
+                    f"Preserved session state: {self.comm_sessions[peer_ip_address[0]][0].ev_session_context}"  # noqa
+                )
+
+        self.tcp_server_handler = None
+        self._current_peer_ip = None
+        if self.udp_server:
+            self.udp_server.resume_udp_server()
+
+    async def start_tcp_server(self, with_tls: bool):
+        if self.tcp_server_handler:
+            logger.info("Reset current tcp handler.")
+            try:
+                await cancel_task(self.tcp_server_handler)
+            except Exception as e:
+                logger.warning(f"Error cancelling existing tcp server handler: {e}")
+            self.tcp_server_handler = None
+
+        server_ready_event: asyncio.Event = asyncio.Event()
+        self.status_event_list.clear()
+        self.status_event_list.append(server_ready_event)
+        if with_tls:
+            self.tcp_server_handler = asyncio.create_task(
+                self.tcp_server.start_tls(server_ready_event)
+            )
+        else:
+            self.tcp_server_handler = asyncio.create_task(
+                self.tcp_server.start_no_tls(server_ready_event)
+            )
+        await self.check_status_task(False)
+
+    async def process_sdp_request(
+        self, sdp_request: SDPRequest
+    ) -> Union[SDPResponse, SDPResponseWireless]:
+        if self.config.enforce_tls or sdp_request.security == Security.TLS:
+            await self.start_tcp_server(True)
+        else:
+            await self.start_tcp_server(False)
+
+        port = self.tcp_server.port
+        # convert IPv6 address from presentation to numeric format
+        ipv6_bytes = socket.inet_pton(
+            socket.AF_INET6, self.tcp_server.ipv6_address_host
+        )
+
+        return create_sdp_response(
+            sdp_request, ipv6_bytes, port, self.tcp_server.is_tls_enabled
+        )
 
     async def process_incoming_udp_packet(self, message: UDPPacketNotification):
         """
@@ -313,58 +420,47 @@ class CommunicationSessionHandler:
         SDP response with or without
         PPD (pairing and positioning device -> ACD-pantograph in ISO 15118-20)
         """
+
         try:
             v2gtp_msg = V2GTPMessage.from_bytes(Protocol.UNKNOWN, message.data)
         except InvalidV2GTPMessageError as exc:
             logger.exception(exc)
             return
 
-        # An incoming datagram can only be an SDP request message, all
-        # other messages are sent via TCP
-        if v2gtp_msg.payload_type == ISOV2PayloadTypes.SDP_REQUEST:
-
-            try:
-                sdp_request = SDPRequest.from_payload(v2gtp_msg.payload)
-                logger.info(f"SDPRequest received: {sdp_request}")
-
-                if self.config.enforce_tls or sdp_request.security == Security.TLS:
-                    port = self.tcp_server.port_tls
-                else:
-                    port = self.tcp_server.port_no_tls
-
-                # convert IPv6 address from presentation to numeric format
-                ipv6_bytes = socket.inet_pton(
-                    socket.AF_INET6, self.tcp_server.ipv6_address_host
+        async with self.udp_processor_lock:
+            # Process one incoming datagram at a time.
+            # An incoming datagram can only be an SDP request message, all
+            # other messages are sent via TCP
+            if v2gtp_msg.payload_type == ISOV2PayloadTypes.SDP_REQUEST:
+                try:
+                    sdp_request = SDPRequest.from_payload(v2gtp_msg.payload)
+                    logger.info(f"SDPRequest received: {sdp_request}")
+                    sdp_response = await self.process_sdp_request(sdp_request)
+                except InvalidSDPRequestError as exc:
+                    logger.exception(
+                        f"{exc.__class__.__name__}, received bytes: "
+                        f"{v2gtp_msg.payload.hex()}"
+                    )
+                    return
+            elif v2gtp_msg.payload_type == ISOV20PayloadTypes.SDP_REQUEST_WIRELESS:
+                raise NotImplementedError(
+                    "The incoming datagram seems to be an SECC Discovery request "
+                    "message for wireless communication (used for ACD-P). "
+                    "This feature is not yet implemented."
                 )
-
-                sdp_response = create_sdp_response(
-                    sdp_request, ipv6_bytes, port, self.config.enforce_tls
-                )
-            except InvalidSDPRequestError as exc:
-                logger.exception(
-                    f"{exc.__class__.__name__}, received bytes: "
-                    f"{v2gtp_msg.payload.hex()}"
+            else:
+                logger.error(
+                    f"Incoming datagram of {len(message.data)} "
+                    f"bytes is no valid SDP request message"
                 )
                 return
-        elif v2gtp_msg.payload_type == ISOV20PayloadTypes.SDP_REQUEST_WIRELESS:
-            raise NotImplementedError(
-                "The incoming datagram seems to be an SECC Discovery request "
-                "message for wireless communication (used for ACD-P). "
-                "This feature is not yet implemented."
-            )
-        else:
-            logger.error(
-                f"Incoming datagram of {len(message.data)} "
-                f"bytes is no valid SDP request message"
-            )
-            return
 
-        # TODO Determine protocol version
-        v2gtp_msg = V2GTPMessage(
-            Protocol.ISO_15118_2,
-            ISOV2PayloadTypes.SDP_RESPONSE,
-            sdp_response.to_payload(),
-        )
-        logger.info(f"Sending SDPResponse: {sdp_response}")
+            # TODO Determine protocol version
+            v2gtp_msg = V2GTPMessage(
+                Protocol.ISO_15118_2,
+                ISOV2PayloadTypes.SDP_RESPONSE,
+                sdp_response.to_payload(),
+            )
+            logger.info(f"Sending SDPResponse: {sdp_response}")
 
-        self.udp_server.send(v2gtp_msg, message.addr)
+            self.udp_server.send(v2gtp_msg, message.addr)
